package page_vote

import (
	"code.electrolab.fr/it/vote.electrolab.fr/app/server"
	"code.electrolab.fr/it/vote.electrolab.fr/app/server/pages/common"
	"code.electrolab.fr/it/vote.electrolab.fr/model"
	"context"
	"fmt"
	"github.com/lib/pq"
	"html/template"
	"log"
	"math/rand"
	"net/http"
	"strings"
	"time"
)

const (
	FORM_VOTER_PRIVATE_TOKEN = "private_token"
)

var (
	templateChoice    = common.MustParseTemplate(TEMPLATE_CHOICE_STR)
	templateScore     = common.MustParseTemplate(TEMPLATE_SCORE_STR)
	templateForbidden = common.MustParseTemplate(TEMPLATE_FORBIDDEN_STR)
)

func RegisterRoutes(mux *http.ServeMux, app *server.App) {
	mux.Handle("/events/Electrolab-AG-2020/votes/", server.NewAppHandler(app, handleGetVote))
}

func handleGetVote(app *server.App, ctx context.Context, w http.ResponseWriter, r *http.Request) {
	var (
		template     *template.Template
		templateVars = new(TemplateVars)
	)

	pathParts := strings.Split(r.URL.Path, "/")
	if len(pathParts) != 5 {
		w.WriteHeader(http.StatusNotFound)
		return
	}

	userUuid, voterFullName, err := common.GetVoterInfo(app, ctx, r)
	if err != nil {
		log.Println(err.Error())
		// Token is not valid
		w.WriteHeader(http.StatusForbidden)
		templateForbidden.Execute(w, nil)
		return
	} else {
		templateVars.VoterPrivateToken = r.FormValue("private_token")
		templateVars.VoterFullname = voterFullName
	}

	questionName := pathParts[4]
	templateVars.Question, err = model.FetchQuestion(app.Context, app.Service.DB, userUuid, questionName)
	if err != nil {
		// Query has an error
		log.Printf("Query error 1: %s", err.Error())
		w.WriteHeader(http.StatusInternalServerError)
		return
	}
	if templateVars.Question == nil {
		// Question is not found
		// w.Header().Set("Location", "/votes")
		w.WriteHeader(http.StatusNotFound)
		return
	}

	templateVars.PageTitle = "Vote Électrolab - " + templateVars.Question.Title

	if r.Method == http.MethodPost {
		err := doVote(app, ctx, r, templateVars)
		if err != nil {
			templateVars.VoteError = err.Error()
			fmt.Println("vote error", templateVars.VoteError)
		} else {
			// Vote is a success, redirect to the page as GET
			w.Header().Set("Location", r.URL.String())
			w.WriteHeader(http.StatusMovedPermanently)
			return
		}
	}

	rows, err := app.Service.DB.QueryContext(ctx, `SELECT name, checked, sentence, vote_count FROM answers WHERE question_name = $1 ORDER BY position`, questionName)
	if err != nil {
		// Query has an error
		log.Printf("Query error 2: %s", err.Error())
		w.WriteHeader(http.StatusInternalServerError)
		return
	}
	defer rows.Close()

	for rows.Next() {
		var answer TemplateAnswer

		err = rows.Scan(&answer.Name, &answer.Checked, &answer.Sentence, &answer.VoteCount)
		if err != nil {
			log.Printf("Query error 3: %s", err.Error())
			w.WriteHeader(http.StatusInternalServerError)
			return
		}

		templateVars.Answers = append(templateVars.Answers, answer)
	}

	// err = app.Service.DB.QueryRowContext(ctx, `SELECT COALESCE((SELECT TRUE FROM votes WHERE question_name = $1 AND user_uuid = $2), FALSE)`, templateVars.QuestionName, userUuid).Scan(&templateVars.VotedAlready)
	// if err != nil {
	// 	// Query has an error
	// 	log.Printf("Query error 7: %s", err.Error())
	// 	w.WriteHeader(http.StatusInternalServerError)
	// 	return
	// }

	if templateVars.Question.VoteOpen || templateVars.Question.VoteCount == 0 {
		template = templateChoice
		shuffleAnswers(templateVars.Answers)
	} else {
		// If the vote is closed, and there is at least one recorded vote, then consider the vote done,
		// and switch to the 'score' template to show the results
		template = templateScore
	}

	err = template.Execute(w, templateVars)
	if err != nil {
		// Template has an error
		log.Printf("Template execution error: %s", err.Error())
		w.WriteHeader(http.StatusInternalServerError)
		return
	}
	return
}

// Pseudorandom shuffle to prevent order bias in answers
func shuffleAnswers(answers []TemplateAnswer) {
	rnd := rand.New(rand.NewSource(time.Now().UnixNano()))
	for i := len(answers) - 1; i > 0; i-- {
		j := rnd.Intn(i + 1)
		answers[i], answers[j] = answers[j], answers[i]
	}
}

func doVote(app *server.App, ctx context.Context, r *http.Request, templateVars *TemplateVars) error {
	var voteErrorMsg string

	err := app.Service.DB.QueryRowContext(ctx,
		"SELECT COALESCE(vote($1, $2, $3), '')",
		templateVars.Question.Name,
		templateVars.VoterPrivateToken,
		pq.StringArray(r.Form["answer"]),
	).Scan(&voteErrorMsg)

	if err != nil {
		return fmt.Errorf("SQL query error: %s", err.Error())
	}

	if len(voteErrorMsg) > 0 {
		return fmt.Errorf("%s", voteErrorMsg)
	}

	return nil
}
